@@ -278,7 +278,7 @@ genPropertyGetter getter n docSection prop = group $ do
 -- | Generate documentation for the given constructor.
 constructorDoc :: Property -> Text
 constructorDoc prop = T.unlines [
-    "Construct a `GValueConstruct` with valid value for the “@" <> propName prop <> "@” property. This is rarely needed directly, but it is used by `Data.GI.Base.Constructible.new`."
+    "Construct a t'GValueConstruct' with valid value for the “@" <> propName prop <> "@” property. This is rarely needed directly, but it is used by `Data.GI.Base.Constructible.new`."
     ]
 
 genPropertyConstructor :: Text -> Name -> HaddockSection -> Property -> ExcCodeGen ()
@@ -448,6 +448,8 @@ genOneProperty owner prop = do
     transferType <- if writable || constructOnly
                     then propTransferType (propType prop)
                     else return "()"
+    let puttable = readable && writable && inConstraint == ("(~) " <> outType)
+
     let allowedOps = (if writable
                       then ["'AttrSet", "'AttrConstruct"]
                       else [])
@@ -460,6 +462,9 @@ genOneProperty owner prop = do
                      <> (if isNullable && propWriteNullable prop /= Just False
                          then ["'AttrClear"]
                          else [])
+                     <> (if puttable
+                         then ["'AttrPut"]
+                         else [])
     it <- infoType owner prop
     export docSection it
     api <- findAPIByName owner
@@ -483,6 +488,9 @@ genOneProperty owner prop = do
             line $ "type AttrOrigin " <> it <> " = " <> name
             line $ "attrGet = " <> getter
             line $ "attrSet = " <> setter
+            if puttable
+              then line $ "attrPut = " <> setter
+              else line $ "attrPut = undefined"
             if writable || constructOnly
               then do line $ "attrTransfer _ v = do"
                       indent $ genPropTransfer "v" (propType prop)
